import java.util.*;

/**
 * <pre>   
 * Topology class keeps track of connections between nodes. 
 * This is a Singleton object
 * </pre>   
 */
public class Topology {
    
    private ArrayList edges;
    private HashMap failedNodes;
    private boolean allToAll;
    private static Topology topology = null;

    /**
     * Static method for getting an instance of Topology
     * @return An instance of Topology
     */
    public static Topology GetInstance() {
	if(Topology.topology == null) {
	    Topology.topology = new Topology(false);	    
	}    
	return Topology.topology;
    }

    /**
     * Static method for getting an instance of Topology
     * @return An instance of Topology
     */
    public static Topology GetInstance(boolean allToAll) {
	if(Topology.topology == null) {
	    Topology.topology = new Topology(allToAll);	    
	}    
	return Topology.topology;
    }

    /**
     * Returns the edge between a and b, if one exists and it is live. Returns null otherwise
     * @param a Int specifying a node
     * @param b Int specifying a node
     * @return The live edge between a and b. If one does not exist then null is returned
     */
    public Edge getLiveEdge(int a, int b) {
	// this code shortcircuits for wireless nodes

	if (withinDist(a, b)) {
	    return getEdge(a, b);
	}
	return null;

	// uncomment the following code for the wired case
// 	Edge e = getEdge(a, b);
// 	if(e == null || !e.isLive() || !this.isNodeAlive(a) || !this.isNodeAlive(b)) {
// 	    return null;
// 	} 
// 	return e;
    }

    /**
     * Returns true if the given node is alive, else return false
     * @param node Int specifying node
     */
    public boolean isNodeAlive(int node) {
	return !this.failedNodes.containsKey(new Integer(node));
    }

    /**
     * Creates a new edge or updates an existing one
     * @param a Int specifying a node
     * @param b Int specifying a node
     * @param options Options to set for the given edge
     */
    public void newEdge(int a, int b, EdgeOptions options) {
	Edge e = this.getEdge(a, b);
	if (e != null) {
	    e.setState(true);
	}else {
	    e = new Edge(a, b, options);
	    edges.add(e);
	}
    }

    /**
     * Mark the given edge as failed so that we don't use it
     * @param a Int specifying a node
     * @param b Int specifying a node
     * @return True if edge was succesfully failed. False if no edge exists between a and b
     */
    public boolean failEdge(int a, int b) {
	return changeEdge(a, b, false);
    }

    /**
     * Restart the given edge so that we can use it again
     * @param a Int specifying a node
     * @param b Int specifying a node
     * @return True if edge was succesfully restarted. False if no edge exists between a and b
     */
    public boolean restartEdge(int a, int b) {
	return changeEdge(a, b, true);
    }
    
    /**
     * Mark the given node as failed, so that we don't use any edge that goes through it
     * @param a Int specifying a node
     */
    public void failNode(int a) {
	this.failedNodes.put(new Integer(a), null);
    }

    /**
     * Mark the given node as ok, so that we can use it again
     * @param a Int specifying a node
     */
    public void restartNode(int a) {
	this.failedNodes.remove(new Integer(a));
    }

    //********* Functions, data structures to deal with mobility **********
    // Range of positions is from 0 to 999.
    private final int REGION_EXTENT = 1000;
    private final int WIRELESS_RADIUS = 300;
    
    private Hashtable ht;
    public void registerMobility(long curr, int a, int xpos, int ypos, int xspeed, int yspeed) {
	moveNodes(curr);
	Integer nodeNum = new Integer(a);
	ht.remove(nodeNum);
	ClientInfo cl = new ClientInfo(a, xpos, ypos, xspeed, yspeed);
	ht.put(nodeNum, cl);
    }

    private long lastTime;
    public void moveNodes(long curr) {
	for (Enumeration e = ht.keys(); e.hasMoreElements(); ) {
	    ClientInfo nd = (ClientInfo)ht.get((Integer)e.nextElement());
	    nd.move((int)(curr - lastTime));
	}
	lastTime = curr;
    }

    public boolean withinDist(int a, int b) {
	ClientInfo na = (ClientInfo)ht.get(new Integer(a));
	ClientInfo nb = (ClientInfo)ht.get(new Integer(b));
	if (na == null || nb == null)
	    return false;
	return na.withinDist(nb, WIRELESS_RADIUS);
    }

    //********** Private Functions **********

    private Topology(boolean allToAll) {
	this.edges = new ArrayList();
	this.failedNodes = new HashMap();
	this.allToAll = allToAll;
	this.ht = new Hashtable();
    }

    private Edge getEdge(int a, int b) {
	ListIterator iter = this.edges.listIterator();
	Edge e;
	while(iter.hasNext()) {
	    e = (Edge)iter.next();
	    if (e.isEdge(a, b)) {
		return e;
	    }
	}
	if(this.allToAll) {
	    // if no edge exists create one
	    e = new Edge(a, b, new EdgeOptions());
	    return e;
	}

	return null;
    }
    
    private boolean changeEdge(int a, int b, boolean state) {
	Edge e = this.getEdge(a, b);
	if(e != null) {
	    e.setState(state);
	    return true;
	}
	
	System.err.println("No edge exists between " + String.valueOf(a) + " and " + String.valueOf(b));	
	return false;
    }

}

class ClientInfo {
    int nodeNum;
    int xpos;
    int ypos;
    int xspeed;
    int yspeed;

    ClientInfo(int node, int xp, int yp, int xsp, int ysp)
    {
	nodeNum = node;
	xpos = xp;
	ypos = yp;
	xspeed = xsp;
	yspeed = ysp;
    }

    String getPos()
    {
	return xpos + " " + ypos;
    }

    void move(int t)
    {
	xpos = (xpos + xspeed*t) % 1000;
	ypos = (ypos + yspeed*t) % 1000;
    }

    boolean withinDist(ClientInfo cl, int rad)
    {
	int x = cl.xpos;
	int y = cl.ypos;
	int xdist, ydist;

	if (x > xpos) {
	    xdist = ((x - xpos) < (1000 + xpos - x) ? (x - xpos) : (1000 + xpos - x));
	} else {
	    xdist = ((xpos - x) < (1000 + x - xpos) ? (xpos - x) : (1000 + x - xpos));
	}
	
	if (y > ypos) {
	    ydist = ((y - ypos) < (1000 + ypos - y) ? (y - ypos) : (1000 + ypos - y));
	} else {
	    ydist = ((ypos - y) < (1000 + y - ypos) ? (ypos - y) : (1000 + y - ypos));
	}

	int distsq = xdist*xdist + ydist*ydist;
	return (distsq < rad*rad);
    }
}
